/*
	SNDXFORM.CPP	- Functions for sample processing

	(c) 1997 Martin Pitt as Public Domain

	Version: 1.0
*/

#include "sndxform.h"
#include <math.h>

void sampleToggleSign (Sample src, unsigned len)
{
  for (unsigned i = 0; i < len; i++)
    src[i] ^= 128;
}

void sampleLoPass (Sample src, Sample dest, unsigned len)
{
  unsigned char last[5] = {128, 128, 128, 128, 128};

  for (unsigned i = 0; i < len; i++)
  {
    last[4] = last[3];
    last[3] = last[2];
    last[2] = last[1];
    last[1] = *last;
    *last = src[i];
    dest[i] = ((unsigned) *last+last[1]+last[2]+last[3]+last[4]) / 5;
  }
}

void sampleHiPass (Sample src, Sample dest, unsigned len)
{
  for (unsigned i = len-1; i; i--)
    dest[i] = src[i]-src[i-1]+128;
  *dest = *src;
}

void sampleMedPass (Sample src, Sample dest, unsigned len,
                   unsigned frq, unsigned midFrq, unsigned width)
{
//* Filterkoeffizienten berechnen **********
  int c = (int) (-100*exp (-2*M_PI*width/frq));
  int b = (int) (200*exp (-M_PI*width/frq)*cos(2*M_PI*midFrq/frq));

  sampleMedPassX (src, dest, len, 100-b-c, b, c);
}

    signed char __filter_clipint (int x)
    {
	if (x < 0)
	    return (x < -128) ? -128 : x;
	else
	    return (x > 127) ? 127 : x;
    }

void sampleMedPassX (Sample src, Sample dest, unsigned len,
                      int a, int b, int c)
{
  signed char *s = (signed char *) src;
  signed char *d = (signed char *) dest;

  sampleToggleSign (src, len);

  *d = __filter_clipint (a*(*s)/100);
  d[1] = __filter_clipint ((a*s[1] + b*(*d))/100);

  for (unsigned i = 2; i < len; i++)
    dest[i] = __filter_clipint((a*s[i] + b*d[i-1] + c*d[i-2])/100);

  sampleToggleSign (src, len);
  sampleToggleSign (dest, len);
}

void fft (float *real, float *imag, unsigned char l)
{
  float scl, arg, c, s, t1, t2;
  int   np, lmx, lo, lix, li, lm, j1, j2, nv2, npm1, i, j, k;

  lmx = np = 1 << l;
  scl = 2*M_PI/np;
  for (lo=1; lo <= l; lo++)
  {
    lix = lmx;
    lmx /= 2;
    arg = 0;
    for (lm = 1; lm <= lmx; lm++)
    {
      c = cos (arg);
      s = sin (arg);
      arg += scl;
      for (li = lix; li <= np; li += lix)
      {
        j1 = li-lix+lm;
        j2 = j1+lmx;
        t1 = real[j1-1] - real[j2-1];
        t2 = imag[j1-1] - imag[j2-1];
        real[j1-1] += real[j2-1];
        imag[j1-1] += imag[j2-1];
        real[j2-1] = c*t1+s*t2;
        imag[j2-1] = c*t2-s*t1;
      }
    }
    scl *= 2;
  }

  j = 1;
  nv2 = np/2;
  npm1 = np-1;
  for (i = 1; i <= npm1; i++)
  {
    if (i < j)
    {
      t1 = real[j-1];
      t2 = imag[j-1];
      real[j-1] = real[i-1];
      imag[j-1] = imag[i-1];
      real[i-1] = t1;
      imag[i-1] = t2;
    }

    k = nv2;
    while (k < j)
    {
      j -= k;
      k /= 2;
    }
    j += k;
  }
}


void fft8 (Sample src, Sample dest)
{
  float *x = new float[256];
  float *y = new float[256];
  unsigned i;

  for (i = 0; i < 256; i++)
  {
    x[i] = (float) src[i] - 128;
    y[i] = 0;
  }

  fft (x, y, 9);

  for (i = 0; i < 256; i++)
    dest[i] = (unsigned) sqrt (x[i]*x[i]+y[i]*y[i]) >> 8;

  delete x;
  delete y;
}

void fft9 (Sample src, Sample dest)
{
  float *x = new float[512];
  float *y = new float[512];
  unsigned i;

  for (i = 0; i < 512; i++)
  {
    x[i] = (float) src[i] - 128;
    y[i] = 0;
  }

  fft (x, y, 9);

  for (i = 0; i < 512; i++)
    dest[i] = (unsigned) sqrt (x[i]*x[i]+y[i]*y[i]) >> 8;

  delete x;
  delete y;
}


void cepstrum (const Sample src, unsigned char l, Sample dest, unsigned firstCfc,
		unsigned numCfc, float amplify)
{
  unsigned srcSize = 1 << l, i;
  float *x = new float[srcSize];
  float *y = new float[srcSize];

  for (i = 0; i < srcSize; i++)
  {
    x[i] = (float) (unsigned char) src[i] - 128;
    y[i] = 0;
  }

  fft (x, y, l);

//* dest := log | x + y*i | = 1/2*log (x^2+y^2) *
  for (i = 0; i < srcSize/2; i++)
  {
    x[i] = sqrt (x[i]*x[i]+y[i]*y[i]) / l;
    x[i] = (x[i] > 1) ? log (x[i]) : 0;
    y[i] = 0;
  }

//* clear (insignificant) second half of fft *
  for (; i < srcSize; i++)
    x[i] = y[i] = 0;

//* calc cepstrum **************************
  fft (x, y, l);

//* calc abs. amount, amplify **************
  for (i = 0; i < numCfc; i++)
      dest[i] = (unsigned char) (sqrt (x[firstCfc+i] * x[firstCfc+i] +
				       y[firstCfc+i] * y[firstCfc+i]) * amplify+0.5);

  delete x;
  delete y;
}
